/***************************************************************************
 * Copyright 2012 by
 *  + Christian-Albrechts-University of Kiel
 *    + Department of Computer Science
 *      + Software Engineering Group 
 *  and others.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ***************************************************************************/

package kieker.webgui.beans.view;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.net.MalformedURLException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.faces.application.FacesMessage;
import javax.faces.application.FacesMessage.Severity;
import javax.faces.bean.ManagedBean;
import javax.faces.bean.ManagedProperty;
import javax.faces.bean.ViewScoped;
import javax.faces.context.FacesContext;

import kieker.analysis.model.analysisMetaModel.MIAnalysisMetaModelFactory;
import kieker.analysis.model.analysisMetaModel.MIDependency;
import kieker.analysis.model.analysisMetaModel.MIDisplay;
import kieker.analysis.model.analysisMetaModel.MIFilter;
import kieker.analysis.model.analysisMetaModel.MIInputPort;
import kieker.analysis.model.analysisMetaModel.MIOutputPort;
import kieker.analysis.model.analysisMetaModel.MIPlugin;
import kieker.analysis.model.analysisMetaModel.MIPort;
import kieker.analysis.model.analysisMetaModel.MIProject;
import kieker.analysis.model.analysisMetaModel.MIProperty;
import kieker.analysis.model.analysisMetaModel.MIRepository;
import kieker.analysis.model.analysisMetaModel.MIRepositoryConnector;
import kieker.analysis.model.analysisMetaModel.impl.MAnalysisMetaModelFactory;
import kieker.analysis.plugin.AbstractPlugin;
import kieker.analysis.plugin.annotation.InputPort;
import kieker.analysis.plugin.annotation.OutputPort;
import kieker.analysis.plugin.annotation.Plugin;
import kieker.analysis.plugin.annotation.RepositoryPort;
import kieker.analysis.plugin.filter.AbstractFilterPlugin;
import kieker.analysis.plugin.reader.AbstractReaderPlugin;
import kieker.analysis.repository.AbstractRepository;
import kieker.analysis.repository.annotation.Repository;
import kieker.common.configuration.Configuration;
import kieker.common.logging.Log;
import kieker.common.logging.LogFactory;
import kieker.monitoring.core.registry.Registry;
import kieker.webgui.beans.application.ProjectsBean;
import kieker.webgui.common.ConnectionFilterToFilter;
import kieker.webgui.common.ConnectionFilterToRepository;
import kieker.webgui.common.FSManager;
import kieker.webgui.common.Global;
import kieker.webgui.common.Pair;
import kieker.webgui.common.PluginFinder;
import kieker.webgui.common.exception.LibraryAlreadyExistingException;
import kieker.webgui.common.exception.NewerProjectException;

import org.primefaces.context.RequestContext;
import org.primefaces.event.FileUploadEvent;
import org.primefaces.model.DefaultTreeNode;
import org.primefaces.model.TreeNode;
import org.primefaces.model.UploadedFile;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;

/**
 * The {@link CurrentAnalysisEditorBean} contains the necessary data behind an instance of the analysis editor. It provides various methods to manipulate the current
 * project, as the analysis editor is the most important part of the whole application.<br>
 * The class is a JSF managed bean with view scope to make sure that one user (even in one session) can open multiple projects at a time without causing any
 * problems.
 * 
 * @author Nils Christian Ehmke
 * @version 1.0
 */
@ManagedBean
@ViewScoped
public final class CurrentAnalysisEditorBean {
	/**
	 * This is the log for errors, exceptions etc.
	 */
	private static final Log LOG = LogFactory.getLog(CurrentAnalysisEditorBean.class);
	/**
	 * This is the factory which will be used to create new components for the project.
	 */
	private final MIAnalysisMetaModelFactory factory = new MAnalysisMetaModelFactory();
	/**
	 * This is the actual model instance. It is the in-memory-model of the current (session) user.
	 */
	private MIProject project;
	/**
	 * This is the corresponding class loader to the project. It contains always the libraries within the lib-folder of the project.
	 */
	private ClassLoader classLoader;
	/**
	 * This is the name of the stored project. It can be used as an identifier within the FS-Manager
	 */
	private String projectName;
	/**
	 * This is the time stamp of the moment, the project was loaded or last saved. It can be used to check whether the project has been modified in the meanwhile.
	 */
	private long timeStamp;
	/**
	 * This list contains the available repositories for the current project. The list is from the type Class<?> to allow a modular algorithm for the tree
	 * assembling.
	 */
	private final List<Class<?>> availableRepositories = Collections.synchronizedList(new ArrayList<Class<?>>());
	/**
	 * This list contains the available filters for the current project. The list is from the type Class<?> to allow a modular algorithm for the tree
	 * assembl
	 */
	private final List<Class<?>> availableFilters = Collections.synchronizedList(new ArrayList<Class<?>>());
	/**
	 * This list contains the available readers for the current project. The list is from the type Class<?> to allow a modular algorithm for the tree
	 * assembl
	 */
	private final List<Class<?>> availableReaders = Collections.synchronizedList(new ArrayList<Class<?>>());
	/**
	 * This map contains the mapping between the plugins and their ID.
	 */
	private Registry<MIPlugin> pluginMap = new Registry<MIPlugin>();
	/**
	 * This map contains the mapping between the ports and their ID.
	 */
	private Registry<MIPort> portMap = new Registry<MIPort>();
	/**
	 * This map contains the mapping between the repositories and their ID.
	 */
	private Registry<MIRepository> repositoryMap = new Registry<MIRepository>();
	/**
	 * This field contains the currently selected plugin.
	 */
	private MIPlugin selectedPlugin;
	/**
	 * The currently selected repository.
	 */
	private MIRepository selectedRepository;
	/**
	 * This list contains the currently available connections between filters.
	 */
	private final List<ConnectionFilterToFilter> filter2filterConnections = new ArrayList<ConnectionFilterToFilter>();
	/**
	 * This list contains the currently available connections between filters and repositories.
	 */
	private final List<ConnectionFilterToRepository> filter2repositoryConnections = new ArrayList<ConnectionFilterToRepository>();

	@ManagedProperty(value = "#{projectsBean}")
	private ProjectsBean projectsBean;

	/**
	 * Creates a new instance of this class.
	 */
	public CurrentAnalysisEditorBean() {
		// No code necessary
	}

	public ProjectsBean getProjectsBean() {
		return this.projectsBean;
	}

	public void setProjectsBean(final ProjectsBean projectsBean) {
		this.projectsBean = projectsBean;
	}

	/**
	 * This method delivers the project stored in this bean.
	 * 
	 * @return The project for this user.
	 */
	public MIProject getProject() {
		synchronized (this) {
			return this.project;
		}
	}

	/**
	 * This method sets the project stored within this bean and returns the new page for the navigation - depending on the given values.
	 * 
	 * @param newName
	 *            The name of the project.
	 */
	public void setProjectName(final String newName) {
		synchronized (this) {
			// Remember the given parameters
			this.projectName = newName;
		}
	}

	/**
	 * This method initializes the bean by using the current project name to load the project. <b>Do not call this method manually. It will only be accessed by
	 * JSF.</b>
	 */
	public void initialize() {
		synchronized (this) {
			// Make sure that the initialization will only be done for the init request.
			if (!FacesContext.getCurrentInstance().isPostback()) {
				this.project = this.projectsBean.openProject(this.projectName);
				this.getConnectionsFromProject();

				if (this.project != null) {
					// Remember the current time! This is important for the later comparison of the time stamps.
					this.resetTimeStamp();
					// Update the class loader
					this.reloadClassLoader();
					// Add the libraries within the lib-folder to the current model
					this.addLibrariesToModel();
					// Load the available readers, filters and repositories
					this.loadToolPalette();
					// Load the hashmaps to get the plugins and ports
					this.intializeHashMaps();
				}
			}
		}
	}

	/**
	 * Assembles and returns the available components as a primefaces tree. Keep in mind that the tree is currently assembled every time this method is called.
	 * 
	 * @return The tree with all available filters, readers and repositories.
	 */
	public TreeNode getAvailableComponentsAsRoot() {
		synchronized (this) {
			// Create the necessary roots
			final TreeNode root = new DefaultTreeNode("Root", null);
			final TreeNode readerRoot = new DefaultTreeNode("Reader", root);
			final TreeNode filterRoot = new DefaultTreeNode("Filter", root);
			final TreeNode repositoryRoot = new DefaultTreeNode("Repository", root);

			// Fill the tree
			CurrentAnalysisEditorBean.assembleTree(readerRoot, this.availableReaders, "readerLeaf");
			CurrentAnalysisEditorBean.assembleTree(filterRoot, this.availableFilters, "filterLeaf");
			CurrentAnalysisEditorBean.assembleTree(repositoryRoot, this.availableRepositories, "repositoryLeaf");

			return root;
		}
	}

	/**
	 * This method is used to "assemble" the tree containing the available components based on the given parameters.
	 * 
	 * @param root
	 *            The root-node used for the tree.
	 * @param clazzList
	 *            The list of available classes which will be used to fill the tree.
	 * @param leafName
	 *            The typename of the leaf-nodes.
	 */
	private static void assembleTree(final TreeNode root, final List<Class<?>> clazzList, final String leafName) {
		final Map<String, TreeNode> map = new HashMap<String, TreeNode>(); // NOPMD (No concurrent access)

		for (final Class<?> clazz : clazzList) {
			// Get the single components of the class name
			final String className = clazz.getName();
			final String[] nameComponents = className.split("\\.");
			final int compLen = nameComponents.length;
			final StringBuilder fullQualifiedName = new StringBuilder();

			TreeNode parent = root;
			for (int i = 0; i < (compLen - 1); i++) {
				fullQualifiedName.append(nameComponents[i]);
				// Get the correct node if it exists already or create it if necessary
				TreeNode node = map.get(fullQualifiedName.toString());

				if (node == null) {
					node = new DefaultTreeNode(nameComponents[i], parent);
					map.put(fullQualifiedName.toString(), node);
				}

				parent = node;
			}
			new DefaultTreeNode(leafName, clazz, parent);
		}
	}

	/**
	 * This method initializes the hash maps of the plugins, the ports and the repositories. The maps are cleaned beforehand.
	 */
	private void intializeHashMaps() {
		synchronized (this) {
			// Clear all maps
			this.pluginMap = new Registry<MIPlugin>();
			this.portMap = new Registry<MIPort>();
			this.repositoryMap = new Registry<MIRepository>();

			// Initialize the plugin map...
			for (final MIPlugin plugin : this.project.getPlugins()) {
				this.pluginMap.get(plugin);
				// ..and the port map
				for (final MIPort port : plugin.getOutputPorts()) {
					this.portMap.get(port);
				}
				if (plugin instanceof MIFilter) {
					for (final MIPort port : ((MIFilter) plugin).getInputPorts()) {
						this.portMap.get(port);
					}
				}
			}
			// Now initialize the repository map
			for (final MIRepository repository : this.project.getRepositories()) {
				this.repositoryMap.get(repository);
			}
		}
	}

	/**
	 * This method loads the list of available readers, filters and repositories, using the current libraries within the model.
	 */
	private void loadToolPalette() {
		synchronized (this) {
			// Clean our tool palette
			this.availableFilters.clear();
			this.availableReaders.clear();
			this.availableRepositories.clear();

			// Make sure there is a project.
			if (this.project != null) {
				// Run through all libraries
				for (final MIDependency lib : this.project.getDependencies()) {
					this.addToToolPalette(lib);
				}
			}
		}
	}

	/**
	 * This method adds all available readers, filters and repositories within the given library to the lists of this bean.
	 * 
	 * @param lib
	 *            The library used to load the plugins and repositories.
	 */
	private void addToToolPalette(final MIDependency lib) {
		synchronized (this) {
			try {
				final List<Class<AbstractRepository>> repositories = PluginFinder.getAllRepositoriesWithinJar(FSManager.getInstance().getURL(lib, this.projectName),
						this.classLoader);
				final List<Class<AbstractPlugin>> plugins = PluginFinder.getAllPluginsWithinJar(FSManager.getInstance().getURL(lib, this.projectName),
						this.classLoader);
				// Now run through the available classes and add all non-abstract classes to our lists
				for (final Class<AbstractRepository> repository : repositories) {
					if (!Modifier.isAbstract(repository.getModifiers())) {
						this.availableRepositories.add(repository);
					}
				}
				for (final Class<? extends AbstractPlugin> plugin : plugins) {
					if (!Modifier.isAbstract(plugin.getModifiers())) {
						// The following cast results in the unchecked-cast-warnings, but we know that the cast should be correct.
						if (AbstractFilterPlugin.class.isAssignableFrom(plugin)) {
							this.availableFilters.add(plugin);
						} else {
							if (AbstractReaderPlugin.class.isAssignableFrom(plugin)) {
								this.availableReaders.add(plugin);
							}
						}
					}
				}
			} catch (final MalformedURLException ex) {
				ex.printStackTrace();
			}
		}
	}

	/**
	 * This method takes all libraries from the lib-folder and adds them to the in-memory-model.
	 */
	private void addLibrariesToModel() {
		synchronized (this) {
			final List<MIDependency> libs = FSManager.getInstance().getModelLibraries(this.projectName);
			// Add them, but remove all existing dependencies so far to avoid double entries. This also makes sure that the model - after it has been opened - points
			// just to valid dependencies (and to all of them).
			this.project.getDependencies().clear();
			this.project.getDependencies().addAll(libs);
		}
	}

	/**
	 * This method reloads the class loader. In other words: The class loader will always be able to load classes from the jar-files within the lib-folder of the
	 * project.
	 */
	private void reloadClassLoader() {
		synchronized (this) {
			this.classLoader = FSManager.getInstance().getClassLoader(this.projectName);
		}
	}

	/**
	 * This method sets the time stamp to the current system time.
	 */
	public void resetTimeStamp() {
		synchronized (this) {
			this.timeStamp = System.currentTimeMillis();
		}
	}

	/**
	 * This method delivers the project name stored in this bean.
	 * 
	 * @return The project name for this user.
	 */
	public String getProjectName() {
		synchronized (this) {
			return this.projectName;
		}
	}

	/**
	 * This method delivers the current time stamp.
	 * 
	 * @return The time stamp for this user.
	 */
	public long getTimeStamp() {
		synchronized (this) {
			return this.timeStamp;
		}
	}

	/**
	 * This method clears the bean. In other words: The stored project is set to null and it will return the page of the project overview for navigation purposes.
	 * 
	 * @return The name of the page of the project overview.
	 */
	public String clearProject() {
		synchronized (this) {
			this.project = null; // NOPMD
			this.projectName = null; // NOPMD
			this.classLoader = null; // NOPMD
			this.selectedPlugin = null; // NOPMD
			this.selectedRepository = null; // NOPMD
			this.timeStamp = 0;
			this.availableFilters.clear();
			this.availableReaders.clear();
			this.availableRepositories.clear();
			this.pluginMap = new Registry<MIPlugin>();
			this.portMap = new Registry<MIPort>();
			this.repositoryMap = new Registry<MIRepository>();
			this.filter2filterConnections.clear();
			this.filter2repositoryConnections.clear();
		}

		return Global.PAGE_PROJECT_OVERVIEW;
	}

	/**
	 * This method can be used to get the description of an {@link AbstractPlugin}- or an {@link AbstractRepository}-class. The description is read via the
	 * annotation.
	 * 
	 * @param clazz
	 *            The class whose description should be extracted.
	 * @return The description for the class or a substitute if none is available. This is in either case human readable.
	 */
	public String getDescription(final Class<?> clazz) {
		// Get the two potential annotations
		final Plugin annotationPlugin = clazz.getAnnotation(Plugin.class);
		final Repository annotationRepository = clazz.getAnnotation(Repository.class);

		// Now check which one of them is available
		if ((annotationPlugin == null) || annotationPlugin.description().isEmpty()) {
			if ((annotationRepository == null) || annotationRepository.description().isEmpty()) {
				// None. Deliver a human readable substitute.
				return "No description available";
			} else {
				return annotationRepository.description();
			}
		} else {
			return annotationPlugin.description();
		}
	}

	public List<InputPort> getInputPorts(final Class<?> clazz) {
		final ArrayList<InputPort> result = new ArrayList<InputPort>();

		for (final Method method : clazz.getMethods()) {
			// Get the potential annotation
			final InputPort annotationPort = method.getAnnotation(InputPort.class);
			// Now check whether it is available
			if (annotationPort != null) {
				result.add(annotationPort);
			}
		}

		return result;
	}

	public List<OutputPort> getOutputPorts(final Class<?> clazz) {
		final ArrayList<OutputPort> result = new ArrayList<OutputPort>();

		// Get the potential annotation
		final Plugin annotationPlugin = clazz.getAnnotation(Plugin.class);

		// Now check whether it is available
		if (annotationPlugin != null) {
			for (final OutputPort oPort : annotationPlugin.outputPorts()) {
				result.add(oPort);
			}
		}

		return result;
	}

	public List<RepositoryPort> getRepositoryPorts(final Class<?> clazz) {
		final ArrayList<RepositoryPort> result = new ArrayList<RepositoryPort>();

		// Get the potential annotation
		final Plugin annotationPlugin = clazz.getAnnotation(Plugin.class);

		// Now check whether it is available
		if (annotationPlugin != null) {
			for (final RepositoryPort rPort : annotationPlugin.repositoryPorts()) {
				result.add(rPort);
			}
		}

		return result;
	}

	/**
	 * This method is the handler for the file upload. It tries to upload the given file and informs the user via the growl-component.
	 * 
	 * @param event
	 *            The upload event.
	 */
	public void handleFileUpload(final FileUploadEvent event) {
		// Get the file from the event
		final UploadedFile file = event.getFile();

		try {
			// Use the file system manager to upload the new file
			final MIDependency lib;
			synchronized (this) {
				lib = FSManager.getInstance().uploadLibrary(file, this.projectName);
				CurrentAnalysisEditorBean.showMessage(FacesMessage.SEVERITY_INFO, "Libary uploaded.");
				// As it seem to have worked, we can add the library to our model.
				this.project.getDependencies().add(lib);
			}
			// Update our class loader and the available plugins & repositories
			this.reloadClassLoader();
			this.addToToolPalette(lib);
		} catch (final LibraryAlreadyExistingException ex) {
			CurrentAnalysisEditorBean.LOG.info("A library with the same name exists already.", ex);
			CurrentAnalysisEditorBean.showMessage(FacesMessage.SEVERITY_WARN, "A library with the same name exists already.");
		} catch (final IOException ex) {
			CurrentAnalysisEditorBean.LOG.error("An error occured while uploading the library.", ex);
			CurrentAnalysisEditorBean.showMessage(FacesMessage.SEVERITY_ERROR, "An error occured while uploading the library.");
		}
	}

	/**
	 * This method delivers the available libraries of this project as a pair of strings. The first element is the name of the library, the second one te size in
	 * MiBytes as human readable string.
	 * 
	 * @return The available libraries.
	 */
	public List<Pair<String, String>> getLibraries() {
		synchronized (this) {
			return FSManager.getInstance().getLibraries(this.projectName);
		}
	}

	/**
	 * This method delivers the available reader-plugins for the current main project. The delivered plugins are never abstract.
	 * 
	 * @return A list with all readers.
	 */
	public final List<Class<?>> getAvailableReaders() {
		return this.availableReaders;
	}

	/**
	 * This method delivers the available filter-plugins for the current main project. The delivered plugins are never abstract.
	 * 
	 * @return A list with all filter.
	 */
	public final List<Class<?>> getAvailableFilters() {
		return this.availableFilters;
	}

	/**
	 * This method delivers the available repositories for the current main project. The delivered repositories are never abstract.
	 * 
	 * @return A list with all repositories.
	 */
	public final List<Class<?>> getAvailableRepositories() {
		return this.availableRepositories;
	}

	/**
	 * This method tries to save the current project and informs the user about success or fail.
	 * 
	 * @param overwriteNewerProject
	 *            This flag determines whether a newer project should be overwritten.
	 */
	public void saveProject(final boolean overwriteNewerProject) {
		synchronized (this) {
			try {
				FSManager.getInstance().saveProject(this.projectName, this.project, this.timeStamp, overwriteNewerProject);
				CurrentAnalysisEditorBean.showMessage(FacesMessage.SEVERITY_INFO, "Project saved.");
				// Update the time stamp!
				this.resetTimeStamp();
			} catch (final IOException ex) {
				CurrentAnalysisEditorBean.LOG.error("An error occured while saving the project.", ex);
				CurrentAnalysisEditorBean.showMessage(FacesMessage.SEVERITY_ERROR, "An error occured while saving the project.");
			} catch (final NewerProjectException ex) {
				CurrentAnalysisEditorBean.LOG.info("The project has been modified externally in the meanwhile.", ex);
				CurrentAnalysisEditorBean.showMessage(FacesMessage.SEVERITY_WARN, "The project has been modified externally in the meanwhile.");
				// Give the user the possibility to force-save the project
				RequestContext.getCurrentInstance().execute("forceSaveDlg.show()");
			}
		}
	}

	/**
	 * This method fills the displays of the given plugin. In other words: It tries to instantiate the given class and to extract the displays. If the instantiation
	 * fails (for various reasons), the method informs the user and executes normally - without an exception.
	 * 
	 * @param clazz
	 *            The class to be used as a base.
	 * @param plugin
	 *            The plugin to be filled.
	 * @return true iff the plugin has been intialized properly.
	 */
	private boolean fillDisplays(final Class<AbstractPlugin> clazz, final MIPlugin plugin) {
		synchronized (this) {
			try {
				// Try to instantiate the given class, using the special constructor of Kieker.
				final AbstractPlugin pluginInstance = clazz.getConstructor(Configuration.class).newInstance(new Configuration());
				// Get the displays and convert them into model instances
				final String[] displayNames = pluginInstance.getAllDisplayNames();
				for (final String displayName : displayNames) {
					final MIDisplay mDisplay = this.factory.createDisplay();
					mDisplay.setName(displayName);
					plugin.getDisplays().add(mDisplay);
				}
				return true;
			} catch (final InstantiationException ex) {
				CurrentAnalysisEditorBean.LOG.error("An error occured while loading the displays of the plugin.", ex);
				CurrentAnalysisEditorBean.showMessage(FacesMessage.SEVERITY_ERROR, "An error occured while loading the displays of the plugin.");
				return false;
			} catch (final IllegalAccessException ex) {
				CurrentAnalysisEditorBean.LOG.error("An error occured while loading the displays of the plugin.", ex);
				CurrentAnalysisEditorBean.showMessage(FacesMessage.SEVERITY_ERROR, "An error occured while loading the displays of the plugin.");
				return false;
			} catch (final InvocationTargetException ex) {
				CurrentAnalysisEditorBean.LOG.error("An error occured while loading the displays of the plugin.", ex);
				CurrentAnalysisEditorBean.showMessage(FacesMessage.SEVERITY_ERROR, "An error occured while loading the displays of the plugin.");
				return false;
			} catch (final NoSuchMethodException ex) {
				CurrentAnalysisEditorBean.LOG.error("An error occured while loading the displays of the plugin.", ex);
				CurrentAnalysisEditorBean.showMessage(FacesMessage.SEVERITY_ERROR, "An error occured while loading the displays of the plugin.");
				return false;
			} catch (final NoClassDefFoundError ex) {
				CurrentAnalysisEditorBean.LOG.error("An error occured while loading the displays of the plugin.", ex);
				CurrentAnalysisEditorBean.showMessage(FacesMessage.SEVERITY_ERROR, "An error occured while loading the displays of the plugin.");
				return false;
			}
		}
	}

	/**
	 * This method fills the ports of the given plugin. In other words: It tries to instantiate the given class and to extract the ports. If the instantiation fails
	 * (for various reasons), the method informs the user and executes normally - without an exception.
	 * 
	 * @param clazz
	 *            The class to be used as a base.
	 * @param plugin
	 *            The plugin to be filled.
	 * @return true iff the plugin has been intialized properly.
	 */
	private boolean fillPorts(final Class<AbstractPlugin> clazz, final MIPlugin plugin) {
		synchronized (this) {
			try {
				// Try to instantiate the given class, using the special constructor of Kieker.
				final AbstractPlugin pluginInstance = clazz.getConstructor(Configuration.class).newInstance(new Configuration());

				// Get the port and use them to initialize the model plugin.
				final String[] inputPortNames = pluginInstance.getAllInputPortNames();
				final String[] outputPortNames = pluginInstance.getAllOutputPortNames();
				final String[] repositoryPortNames = pluginInstance.getAllRepositoryPortNames();

				// Add input ports
				if (plugin instanceof MIFilter) {
					for (final String inputPortName : inputPortNames) {
						final MIInputPort mInputPort = this.factory.createInputPort();
						mInputPort.setName(inputPortName);

						mInputPort.setParent((MIFilter) plugin);
					}
				}

				// Add output ports.
				for (final String outputPortName : outputPortNames) {
					final MIOutputPort mOutputPort = this.factory.createOutputPort();
					mOutputPort.setName(outputPortName);
					mOutputPort.setParent(plugin);
				}

				// Add repository ports.
				for (final String repositoryPortName : repositoryPortNames) {
					final MIRepositoryConnector mConnector = this.factory.createRepositoryConnector();
					mConnector.setName(repositoryPortName);
					plugin.getRepositories().add(mConnector);
				}

				return true;

			} catch (final InstantiationException ex) {
				CurrentAnalysisEditorBean.LOG.error("An errcor occured while loading the ports of the plugin.", ex);
				CurrentAnalysisEditorBean.showMessage(FacesMessage.SEVERITY_ERROR, "An errcor occured while loading the ports of the plugin.");
				return false;
			} catch (final IllegalAccessException ex) {
				CurrentAnalysisEditorBean.LOG.error("An errcor occured while loading the ports of the plugin.", ex);
				CurrentAnalysisEditorBean.showMessage(FacesMessage.SEVERITY_ERROR, "An errcor occured while loading the ports of the plugin.");
				return false;
			} catch (final InvocationTargetException ex) {
				CurrentAnalysisEditorBean.LOG.error("An errcor occured while loading the ports of the plugin.", ex);
				CurrentAnalysisEditorBean.showMessage(FacesMessage.SEVERITY_ERROR, "An errcor occured while loading the ports of the plugin.");
				return false;
			} catch (final NoSuchMethodException ex) {
				CurrentAnalysisEditorBean.LOG.error("An errcor occured while loading the ports of the plugin.", ex);
				CurrentAnalysisEditorBean.showMessage(FacesMessage.SEVERITY_ERROR, "An errcor occured while loading the ports of the plugin.");
				return false;
			} catch (final NoClassDefFoundError ex) {
				CurrentAnalysisEditorBean.LOG.error("An errcor occured while loading the ports of the plugin.", ex);
				CurrentAnalysisEditorBean.showMessage(FacesMessage.SEVERITY_ERROR, "An errcor occured while loading the ports of the plugin.");
				return false;
			}
		}
	}

	/**
	 * This method fills the properties of the given repository. In other words: It tries to instantiate the given class and to extract the configuration keys. If
	 * the instantiation fails (for various reasons), the method informs the user and executes normally - without an exception.
	 * 
	 * @param clazz
	 *            The class to be used as a base.
	 * @param repository
	 *            The repository to be filled.
	 * @return true iff the repository has been intialized properly.
	 */
	private boolean fillProperties(final Class<AbstractRepository> clazz, final MIRepository repository) {
		try {
			// Try to instantiate the given class, using the special constructor of Kieker.
			final AbstractRepository repositoryInstance = clazz.getConstructor(Configuration.class).newInstance(new Configuration());
			// Get the current configuration and use it to initialize the model repository, as THIS configuration instance will contain all keys.
			final Configuration configuration = repositoryInstance.getCurrentConfiguration();

			repository.getProperties().addAll(this.extractProperties(configuration));

			return true;
		} catch (final InstantiationException ex) {
			CurrentAnalysisEditorBean.LOG.error("An errcor occured while loading the properties of the repository.", ex);
			CurrentAnalysisEditorBean.showMessage(FacesMessage.SEVERITY_ERROR, "An errcor occured while loading the properties of the repository.");
			return false;
		} catch (final IllegalAccessException ex) {
			CurrentAnalysisEditorBean.LOG.error("An errcor occured while loading the properties of the repository.", ex);
			CurrentAnalysisEditorBean.showMessage(FacesMessage.SEVERITY_ERROR, "An errcor occured while loading the properties of the repository.");
			return false;
		} catch (final InvocationTargetException ex) {
			CurrentAnalysisEditorBean.LOG.error("An errcor occured while loading the properties of the repository.", ex);
			CurrentAnalysisEditorBean.showMessage(FacesMessage.SEVERITY_ERROR, "An errcor occured while loading the properties of the repository.");
			return false;
		} catch (final NoSuchMethodException ex) {
			CurrentAnalysisEditorBean.LOG.error("An errcor occured while loading the properties of the repository.", ex);
			CurrentAnalysisEditorBean.showMessage(FacesMessage.SEVERITY_ERROR, "An errcor occured while loading the properties of the repository.");
			return false;
		} catch (final NoClassDefFoundError ex) {
			CurrentAnalysisEditorBean.LOG.error("An errcor occured while loading the properties of the repository.", ex);
			CurrentAnalysisEditorBean.showMessage(FacesMessage.SEVERITY_ERROR, "An errcor occured while loading the properties of the repository.");
			return false;
		}
	}

	/**
	 * This method fills the properties of the given plugin. In other words: It tries to instantiate the given class and to extract the configuration keys. If
	 * the instantiation fails (for various reasons), the method informs the user and executes normally - without an exception.
	 * 
	 * @param clazz
	 *            The class to be used as a base.
	 * @param plugin
	 *            The plugin to be filled.
	 * @return true iff the plugin has been intialized properly.
	 */
	private boolean fillProperties(final Class<AbstractPlugin> clazz, final MIPlugin plugin) {
		try {
			// Try to instantiate the given class, using the special constructor of Kieker.
			final AbstractPlugin pluginInstance = clazz.getConstructor(Configuration.class).newInstance(new Configuration());
			// Get the current configuration and use it to initialize the model plugin, as THIS configuration instance will contain all keys.
			final Configuration configuration = pluginInstance.getCurrentConfiguration();

			plugin.getProperties().addAll(this.extractProperties(configuration));

			return true;
		} catch (final InstantiationException ex) {
			CurrentAnalysisEditorBean.LOG.error("An errcor occured while loading the properties of the plugin.", ex);
			CurrentAnalysisEditorBean.showMessage(FacesMessage.SEVERITY_ERROR, "An errcor occured while loading the properties of the plugin.");
			return false;
		} catch (final IllegalAccessException ex) {
			CurrentAnalysisEditorBean.LOG.error("An errcor occured while loading the properties of the plugin.", ex);
			CurrentAnalysisEditorBean.showMessage(FacesMessage.SEVERITY_ERROR, "An errcor occured while loading the properties of the plugin.");
			return false;
		} catch (final InvocationTargetException ex) {
			CurrentAnalysisEditorBean.LOG.error("An errcor occured while loading the properties of the plugin.", ex);
			CurrentAnalysisEditorBean.showMessage(FacesMessage.SEVERITY_ERROR, "An errcor occured while loading the properties of the plugin.");
			return false;
		} catch (final NoSuchMethodException ex) {
			CurrentAnalysisEditorBean.LOG.error("An errcor occured while loading the properties of the plugin.", ex);
			CurrentAnalysisEditorBean.showMessage(FacesMessage.SEVERITY_ERROR, "An errcor occured while loading the properties of the plugin.");
			return false;
		} catch (final NoClassDefFoundError ex) {
			CurrentAnalysisEditorBean.LOG.error("An errcor occured while loading the properties of the plugin.", ex);
			CurrentAnalysisEditorBean.showMessage(FacesMessage.SEVERITY_ERROR, "An errcor occured while loading the properties of the plugin.");
			return false;
		}
	}

	/**
	 * This method extracts the properties from the given configuration. In other words: For every key within the configuration, the method creates a model
	 * counterpart.
	 * 
	 * @param configuration
	 *            The configuration to be used for extraction.
	 * @return The list containing one {@link MIProperty} for every key within the configuration.
	 */
	private List<MIProperty> extractProperties(final Configuration configuration) {
		final List<MIProperty> result = new ArrayList<MIProperty>();

		// Run through all entries.
		final Iterator<Map.Entry<Object, Object>> iterator = configuration.entrySet().iterator();
		while (iterator.hasNext()) {
			final Map.Entry<Object, Object> entry = iterator.next();
			// Create a property object for the current entry.
			final MIProperty property = this.factory.createProperty();
			property.setName(entry.getKey().toString());
			property.setValue(entry.getValue().toString());

			result.add(property);
		}

		return result;
	}

	/**
	 * This method adds a new repository to the current model, using the given instance of {@code Class} for it.
	 * 
	 * @param clazz
	 *            The class of the repository to be added.
	 */
	public void addRepository(final Class<AbstractRepository> clazz) {
		// Create a new instance for the model
		final MIRepository repository = this.factory.createRepository();
		repository.setClassname(clazz.getName());
		repository.setName(clazz.getSimpleName());

		final boolean check = this.fillProperties(clazz, repository);
		// Make sure that everything has been loaded correctly. If something went wrong, we don't add the component to our model
		if (check) {
			synchronized (this) {
				// Add it to the project
				this.project.getRepositories().add(repository);
				this.repositoryMap.get(repository);
			}
		}
	}

	/**
	 * This method adds a new plugin to the current model, using the given instance of {@code Class} for it.
	 * 
	 * @param clazz
	 *            The class of the plugin to be added. This can be both, a filter or a reader.
	 */
	public void addPlugin(final Class<AbstractPlugin> clazz) {
		// Create a new instance for the model
		final MIPlugin plugin;
		if (AbstractReaderPlugin.class.isAssignableFrom(clazz)) {
			plugin = this.factory.createReader();
		} else {
			plugin = this.factory.createFilter();
		}
		plugin.setClassname(clazz.getName());
		plugin.setName(clazz.getSimpleName());

		boolean check = true;
		check &= this.fillProperties(clazz, plugin);
		check &= this.fillPorts(clazz, plugin);
		check &= this.fillDisplays(clazz, plugin);

		// Make sure that everything has been loaded correctly. If something went wrong, we don't add the component to our model
		if (check) {
			synchronized (this) {
				// Add it to the project and don't forget to add the ports.
				this.project.getPlugins().add(plugin);
				this.pluginMap.get(plugin);
				for (final MIPort port : plugin.getOutputPorts()) {
					this.portMap.get(port);
				}
				if (plugin instanceof MIFilter) {
					for (final MIPort port : ((MIFilter) plugin).getInputPorts()) {
						this.portMap.get(port);
					}
				}

				// Make sure the user can edit the repository connections.
				final EList<MIRepositoryConnector> mConnectors = plugin.getRepositories();
				for (final MIRepositoryConnector mConnector : mConnectors) {
					this.filter2repositoryConnections.add(new ConnectionFilterToRepository(plugin, mConnector.getRepository(), mConnector));
				}
			}
		}
	}

	/**
	 * This method removes the given repository from the project. If it is also the currently selected repository, it will be deselected.
	 * 
	 * @param repository
	 *            The repository to be removed from the project.
	 */
	public void removeRepository(final MIRepository repository) {
		synchronized (this) {
			this.project.getRepositories().remove(repository);

			// Remove the corresponding repositories from the connections
			final List<ConnectionFilterToRepository> fRDelList = new ArrayList<ConnectionFilterToRepository>();
			for (final ConnectionFilterToRepository conn : this.filter2repositoryConnections) {
				if (conn.getDestination() == repository) {
					fRDelList.add(conn);
				}
			}

			// Remove them from the project as well
			for (final ConnectionFilterToRepository conn : fRDelList) {
				conn.getOutputPort().setRepository(null);
			}

			this.repositoryMap.remove(repository);

			if (this.selectedRepository == repository) {
				this.selectedRepository = null; // NOPMD
			}
		}
	}

	/**
	 * Uses the given plugin to remove the corresponding connections (to other plugins or repositories) from the current project.
	 * 
	 * @param plugin
	 *            The plugin whose connections have to be removed.
	 */
	private void removeCorrespondingConnections(final MIPlugin plugin) {
		synchronized (this) {
			final List<ConnectionFilterToFilter> ffDelList = new ArrayList<ConnectionFilterToFilter>();
			for (final ConnectionFilterToFilter conn : this.filter2filterConnections) {
				if ((conn.getSource() == plugin) || (conn.getDestination() == plugin)) {
					ffDelList.add(conn);
				}
			}
			this.filter2filterConnections.removeAll(ffDelList);

			// Remove them from the project as well
			for (final ConnectionFilterToFilter conn : ffDelList) {
				if (conn.getDestination() == plugin) {
					conn.getOutputPort().getSubscribers().remove(conn.getInputPort());
				}
			}

			final List<ConnectionFilterToRepository> fRDelList = new ArrayList<ConnectionFilterToRepository>();
			for (final ConnectionFilterToRepository conn : this.filter2repositoryConnections) {
				if (conn.getSource() == plugin) {
					fRDelList.add(conn);
				}
			}
			this.filter2repositoryConnections.removeAll(fRDelList);
		}
	}

	/**
	 * Uses the given plugin to remove the corresponding ports from the current project.
	 * 
	 * @param plugin
	 *            The plugin whose ports have to be removed.
	 */
	private void removeCorrespondingPorts(final MIPlugin plugin) {
		synchronized (this) {
			// Remove the ports from the registry
			for (final MIPort mPort : plugin.getOutputPorts()) {
				this.portMap.remove(mPort);
			}
			// Remove the ports from the project
			if (plugin instanceof MIFilter) {
				for (final MIPort mPort : ((MIFilter) plugin).getInputPorts()) {
					this.portMap.remove(mPort);
				}
			}
		}
	}

	/**
	 * This method removes the given plugin from the project. If it is also the currently selected plugin, it will be deselected.
	 * 
	 * @param plugin
	 *            The plugin to be removed from the project.
	 */
	public void removePlugin(final MIPlugin plugin) {
		synchronized (this) {
			this.project.getPlugins().remove(plugin);
			this.pluginMap.remove(plugin);

			this.removeCorrespondingConnections(plugin);
			this.removeCorrespondingPorts(plugin);

			if (this.selectedPlugin == plugin) {
				this.selectedPlugin = null; // NOPMD
			}
		}
	}

	/**
	 * This method delivers the currently selected plugin (or the currently selected repository). If nothing has been selected, null will be returned.
	 * 
	 * @return The currently selected plugin or repository.
	 */
	public EObject getSelectedPlugin() {
		synchronized (this) {
			if (this.selectedPlugin != null) {
				return this.selectedPlugin;
			} else {
				return this.selectedRepository;
			}
		}
	}

	/**
	 * This method sets the currently selected plugin. The selected repository is set automatically to null.
	 * 
	 * @param selectedPlugin
	 *            The new selection.
	 */
	public void setSelectedPlugin(final MIPlugin selectedPlugin) {
		synchronized (this) {
			this.selectedPlugin = selectedPlugin;
			this.selectedRepository = null; // NOPMD
		}
	}

	/**
	 * This method sets the currently selected repository. The selected plugin is set automatically to null.
	 * 
	 * @param selectedRepository
	 *            The new selection.
	 */
	public void setSelectedRepository(final MIRepository selectedRepository) {
		synchronized (this) {
			this.selectedRepository = selectedRepository;
			this.selectedPlugin = null; // NOPMD
		}
	}

	/**
	 * This method delivers the properties of the currently selected plugin, but it adds also the name- and the class-properties as a string to the list. The first
	 * element is always the classname, the second one is always the pluginname.
	 * 
	 * @return A list with all properties of the plugin plus the name- and class-properties. If no plugin is selected, the list of the selected repository is
	 *         delivered.
	 */
	public List<Object> getAdvancedPluginProperties() {
		synchronized (this) {
			final List<Object> result = new ArrayList<Object>();

			// Add the properties as strings
			result.add("Name");
			result.add("ClassName");

			// Get the original properties of the plugin
			if (this.selectedPlugin != null) {
				result.addAll(this.selectedPlugin.getProperties());
			} else {
				result.addAll(this.selectedRepository.getProperties());
			}

			return result;
		}
	}

	/**
	 * This method delivers a list containing all available filters within the model instance. It is the same as calling project.getPlugins() but instead this method
	 * returns only instances of {@link MIFilter}.
	 * 
	 * @return A list with the available filters.
	 */
	public List<MIPlugin> getFilters() {
		synchronized (this) {
			final EList<MIPlugin> plugins = this.project.getPlugins();
			final List<MIPlugin> filter = new ArrayList<MIPlugin>();

			for (final MIPlugin plugin : plugins) {
				if (plugin instanceof MIFilter) {
					filter.add(plugin);
				}
			}

			return filter;
		}
	}

	/**
	 * This method extracts the connections between the filters from the current main project.
	 */
	private void getConnectionsFromProject() {
		synchronized (this) {
			this.filter2filterConnections.clear();
			this.filter2repositoryConnections.clear();

			if (this.project != null) {
				final EList<MIPlugin> mPlugins = this.project.getPlugins();

				for (final MIPlugin mPlugin : mPlugins) {
					final EList<MIOutputPort> mOutputPorts = mPlugin.getOutputPorts();
					for (final MIOutputPort mOutputPort : mOutputPorts) {
						final EList<MIInputPort> mInputPorts = mOutputPort.getSubscribers();
						for (final MIInputPort mInputPort : mInputPorts) {
							this.filter2filterConnections.add(new ConnectionFilterToFilter(mPlugin, mInputPort.getParent(), mInputPort, mOutputPort));
						}
					}

					final EList<MIRepositoryConnector> mConnectors = mPlugin.getRepositories();
					for (final MIRepositoryConnector mConnector : mConnectors) {
						this.filter2repositoryConnections.add(new ConnectionFilterToRepository(mPlugin, mConnector.getRepository(), mConnector));
					}
				}
			}
		}
	}

	/**
	 * Delivers the connections (between the filters) within the current main project.
	 * 
	 * @return A list containing all available connections.
	 */
	public List<ConnectionFilterToFilter> getFilterConnections() {
		return this.filter2filterConnections;
	}

	/**
	 * Delivers the connections (between filters and repositories) within the current main project.
	 * 
	 * @return A list containing all available connections.
	 */
	public List<ConnectionFilterToRepository> getRepoConnections() {
		return this.filter2repositoryConnections;
	}

	/**
	 * Delivers the <b>valid</b> connections (between filters and repositories) within the current main project.
	 * 
	 * @return A list containing all available and valid connections.
	 */
	public List<ConnectionFilterToRepository> getValidRepoConnections() {
		final List<ConnectionFilterToRepository> list = new ArrayList<ConnectionFilterToRepository>();

		synchronized (this) {
			for (final ConnectionFilterToRepository conn : this.filter2repositoryConnections) {
				if (conn.isValid()) {
					list.add(conn);
				}
			}
		}

		return list;
	}

	/**
	 * Delivers all available repositories.
	 * 
	 * @return A list with all repositories within the model.
	 */
	public EList<MIRepository> getRepositories() {
		synchronized (this) {
			return this.project.getRepositories();
		}
	}

	/**
	 * Delivers the name of the currently selected plugin/repository. This is only necessary for the correct renaming of the components.
	 * 
	 * @return The name of the plugin/repository.
	 */
	public String getCurrentPluginName() {
		synchronized (this) {
			if (this.selectedPlugin != null) {
				return this.selectedPlugin.getName();
			} else {
				return this.selectedRepository.getName();
			}
		}
	}

	/**
	 * Delivers the valid connections (between the filters) within the current main project.
	 * 
	 * @return A list containing all available and valid connections.
	 */
	public List<ConnectionFilterToFilter> getValidConnections() {
		synchronized (this) {
			final List<ConnectionFilterToFilter> validConnections = new ArrayList<ConnectionFilterToFilter>();
			final List<ConnectionFilterToFilter> availableConnections = this.getFilterConnections();

			for (final ConnectionFilterToFilter connection : availableConnections) {
				if (connection.isValid()) {
					validConnections.add(connection);
				}
			}

			return validConnections;
		}
	}

	/**
	 * This method adds an empty connection to the current main project.
	 */
	public void addConnection() {
		this.filter2filterConnections.add(new ConnectionFilterToFilter(null, null, null, null));
	}

	/**
	 * This method "submits" the current connections to the main project (In other words: The connections will be stored within the main project).
	 */
	public void submitConnections() {
		synchronized (this) {
			for (final ConnectionFilterToFilter connection : this.filter2filterConnections) {
				if (connection.isValid()) {
					connection.getOutputPort().getSubscribers().add(connection.getInputPort());
				}
			}

			for (final ConnectionFilterToRepository connection : this.filter2repositoryConnections) {
				if (connection.isValid()) {
					connection.getOutputPort().setRepository(connection.getDestination());
				}
			}
		}
	}

	/**
	 * This method shows the current user a message by using the growl-component of PrimeFaces.
	 * 
	 * @param severity
	 *            The severity of the message.
	 * @param msg
	 *            The message itself.
	 */
	private static void showMessage(final Severity severity, final String msg) {
		FacesContext.getCurrentInstance().addMessage(null, new FacesMessage(severity, "", msg));
	}

	/**
	 * Delivers the id which the plugin has in the plugin map of this bean.
	 * 
	 * @param plugin
	 *            The plugin whose id should be delivered.
	 * @return The id of the plugin.
	 */
	public int getPluginID(final MIPlugin plugin) {
		return this.pluginMap.get(plugin);
	}

	/**
	 * Delivers the id which the repository has in the repository map of this bean.
	 * 
	 * @param repository
	 *            The repository whose id should be delivered.
	 * @return The id of the repository.
	 */
	public int getRepositoryID(final MIRepository repository) {
		return this.repositoryMap.get(repository);
	}

	/**
	 * Delivers the id which the port has in the port map of this bean.
	 * 
	 * @param port
	 *            The port whose id should be delivered.
	 * @return The id of the port.
	 */
	public int getPortID(final MIPort port) {
		return this.portMap.get(port);
	}

	/**
	 * This method searches a plugin by name (the name should be delivered using the toString-method of the object).
	 * 
	 * @param string
	 *            The name of the plugin.
	 * @return The plugin with the given name if it exists, null otherwise.
	 */
	public MIPlugin getPluginByID(final String string) {
		return this.pluginMap.get(Integer.parseInt(string));
	}

	/**
	 * This method searches a port by name (the name should be delivered using the toString-method of the object).
	 * 
	 * @param string
	 *            The name of the port.
	 * @return The port with the given name if it exists, null otherwise.
	 */
	public MIPort getPortByID(final String string) {
		return this.portMap.get(Integer.parseInt(string));
	}

	/**
	 * This method searches a repository by name (the name should be delivered using the toString-method of the object).
	 * 
	 * @param string
	 *            The name of the repository.
	 * @return The plugin with the given name if it exists, null otherwise.
	 */
	public MIRepository getRepositoryByID(final String string) {
		return this.repositoryMap.get(Integer.parseInt(string));
	}
}
